// BEPU template v0.01    02/11/2015

// A basic project including many or all BEPU engine settings
// and an example scene with a ground and a box + the base camera from BEPUDemos

using System;
using System.Diagnostics;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

using BEPUphysics;
using BEPUutilities.Threading;
using BEPUphysicsDrawer.Models;
using BEPUphysics.Entities.Prefabs;
using BEPUphysics.Entities;
using BEPUphysics.PositionUpdating;
using BEPUphysics.CollisionShapes.ConvexShapes;

using BEPUphysics.Constraints.SolverGroups;
using BEPUphysics.CollisionShapes;
using System.Collections.Generic;
using BEPUphysics.Settings;
using BEPUphysics.Constraints;


namespace $safeprojectname$
{
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;

        private ParallelLooper parallelLooper;

        public Camera Camera;
        public ModelDrawer ModelDrawer;

        public KeyboardState KeyboardInput;
        public KeyboardState PreviousKeyboardInput;
        public MouseState MouseInput;
        public MouseState PreviousMouseInput;

        public Space Space;

        Entity kapow;
        Entity ground;
        Entity demo;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            parallelLooper = new ParallelLooper();

            if (Environment.ProcessorCount > 1)
            {
                for (int i = 0; i < Environment.ProcessorCount; i++)
                {
                    parallelLooper.AddThread();
                }
            }

            Space = new Space(parallelLooper);
            //Space = new Space();

            // BEPU fine tuning
            //Space.Solver.IterationLimit = 100;
            //CollisionResponseSettings.Softness = 0f;
            //CollisionDetectionSettings.AllowedPenetration = 0.002f;
            //SolverSettings.DefaultMinimumImpulse = 0f;
            //BEPUutilities.Toolbox.Epsilon = 0.002f;
            //BEPUutilities.Toolbox.BigEpsilon = 0.002f;
            //InertiaHelper.InertiaTensorScale = 1.0f;

            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;

            Camera = new Camera(new BEPUutilities.Vector3(0,20,40), 0, 0, BEPUutilities.Matrix.CreatePerspectiveFieldOfViewRH(BEPUutilities.MathHelper.PiOver4, graphics.PreferredBackBufferWidth / (float)graphics.PreferredBackBufferHeight, .1f, 10000));

            Camera.LockedUp = BEPUutilities.Vector3.Up;
            Camera.ViewDirection = new BEPUutilities.Vector3(0, 0, -1);

            kapow = new Sphere(new BEPUutilities.Vector3(11000, 0, 0), .6f, 20);
            //kapow.Material
            Space.Add(kapow);

            Space.ForceUpdater.Gravity = new BEPUutilities.Vector3(0, -9.81f, 0f);
        }


        protected override void Initialize()
        {
            ModelDrawer = new InstancedModelDrawer(this);
            ModelDrawer.Clear();

            example();

            base.Initialize();
        }

        protected override void LoadContent()
        {
            IsFixedTimeStep = false;

            Mouse.SetPosition(200, 200);

            GC.Collect();
        }

        protected override void UnloadContent()
        {
        }

        protected override void Update(GameTime gameTime)
        {
            PreviousKeyboardInput = KeyboardInput;
            KeyboardInput = Keyboard.GetState();
            PreviousMouseInput = MouseInput;
            MouseInput = Mouse.GetState();

            var dt = (float)gameTime.ElapsedGameTime.TotalSeconds;

            if (KeyboardInput.IsKeyDown(Keys.Escape))
            {
                parallelLooper.Dispose();

                Exit();
            }

            if (!IsMouseVisible)
                Mouse.SetPosition(200, 200);

            Camera.Yaw((200 - MouseInput.X) * dt * 0.012f);
            Camera.Pitch((200 - MouseInput.Y) * dt * 0.012f);

            float Speed = 10f;
            float distance = Speed * dt;

            if (KeyboardInput.IsKeyDown(Keys.E))
                Camera.MoveForward(distance);
            if (KeyboardInput.IsKeyDown(Keys.D))
                Camera.MoveForward(-distance);
            if (KeyboardInput.IsKeyDown(Keys.S))
                Camera.MoveRight(-distance);
            if (KeyboardInput.IsKeyDown(Keys.F))
                Camera.MoveRight(distance);
            if (KeyboardInput.IsKeyDown(Keys.A))
                Camera.MoveUp(distance);
            if (KeyboardInput.IsKeyDown(Keys.Z))
                Camera.MoveUp(-distance);

            if (MouseInput.LeftButton == ButtonState.Pressed)
            {
                kapow.Position = Camera.Position + Camera.WorldMatrix.Forward;
                kapow.AngularVelocity = BEPUutilities.Vector3.Zero;
                kapow.PositionUpdateMode = PositionUpdateMode.Continuous;
                kapow.LinearVelocity = Camera.WorldMatrix.Forward * 50;

                ModelDrawer.Add(kapow);
            }


            if (MouseInput.XButton1 == Microsoft.Xna.Framework.Input.ButtonState.Pressed)
            {
                //Interpolation isn't used in the demos by default, so passing in a really short time adds a lot of time between discretely visible time steps.
                //Using a Space.TimeStepSettings.TimeStepDuration of 1/60f (the default), this will perform one time step every 20 frames (about three per second at the usual game update rate).
                //This can make it easier to examine behavior frame-by-frame.
                Space.Update(1 / 1200f);
            }
            else
            {
                Space.Update();
            }

            ModelDrawer.Update();

            PreviousKeyboardInput = KeyboardInput;
            PreviousMouseInput = MouseInput;

            base.Update(gameTime);
        }

        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(new Color(.41f, .41f, .45f, 1));

            var viewMatrix = Camera.ViewMatrix;
            var projectionMatrix = Camera.ProjectionMatrix;

            ModelDrawer.Draw(viewMatrix, projectionMatrix);

            base.Draw(gameTime);
        }


        private void example()
        {
            ground = new Entity(new BoxShape(500, 1, 500)) { Position = BEPUutilities.Vector3.Zero };
            ModelDrawer.Add(ground);
            Space.Add(ground);

            demo = new Entity(new BoxShape(10, 10, 10), 5) { Position = new BEPUutilities.Vector3(0, 10, 0) };
            Space.Add(demo);
            ModelDrawer.Add(demo);
        }
    }
}